import numpy as np
import scvelo as scv

# Visualization Libraries
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import seaborn as sns


def plot_streamline(adata, args):
    scv.tl.velocity_graph(adata, vkey='sde_velocity', n_jobs=args.scv_n_jobs)
    scv.tl.velocity_confidence(adata, vkey='sde_velocity')
    plt.rcParams.update({'figure.max_open_warning': 0})
    scv.pl.velocity_embedding_stream(adata, 
                                          legend_loc='right_margin', 
                                          vkey="sde_velocity", 
                                          basis=args.vis_key, 
                                          color=[args.vis_type_col], 
                                          title=f"SDEvelo_seed={args.seed}_Epoch={args.nEpochs}", 
                                          cmap="Blues",
                                          figsize=(4, 3), 
                                          dpi=300, 
                                          show=True) 
    
def plot_latent_time(adata, args, figsize=(4.5, 3), dpi=300):
    plt.rcParams.update({'figure.max_open_warning': 0})
    plt.figure(figsize=figsize, dpi=dpi)
    ax = sns.scatterplot(x=adata.obsm[args.vis_key][:, 0], y=adata.obsm[args.vis_key][:, 1],
                         c=adata.obs['latent_time'], cmap="viridis_r")
    norm = plt.Normalize(0, 1)
    sm = plt.cm.ScalarMappable(cmap="viridis_r", norm=norm)
    sm.set_array([])

    # Remove the legend and add a colorbar
    ax.collections[0].set_sizes([150])
    plt.colorbar(sm, ax=ax)
    plt.axis('off')
    plt.show()


def plot_noise_histogram(adata, bins=50, figsize=(6, 3), dpi=300):
    """
    Plot a stacked histogram of sigma1 and sigma2 from the SDEvelo model.

    - param model: The model object containing sigma1 and sigma2 as attributes.
    - param bins: int, the number of bins for the histogram.
    - param figsize: tuple of int, the size of the figure (width, height).
    - param dpi: int, the resolution in dots per inch.
    """
    # Extract sigma1 and sigma2 from the model
    sigma1 = adata.var['fit_sigma_1']
    sigma2 = adata.var['fit_sigma_2']

    # Define the common range for both datasets
    min_bin = min(min(sigma1), min(sigma2))
    max_bin = max(max(sigma1), max(sigma2))
    common_bins = np.linspace(min_bin, max_bin, bins)

    # Create the histogram
    plt.rcParams.update({'figure.max_open_warning': 0})
    plt.figure(figsize=figsize, dpi=dpi)
    plt.hist([sigma1, sigma2], bins=common_bins, stacked=True, color=['blue', 'red'], alpha=0.7)

    plt.xlabel('Value')
    plt.ylabel('Frequency')
    plt.legend(['Sigma1', 'Sigma2'])

    # Save the figure
    # plt.savefig(filename, dpi=dpi, bbox_inches='tight')
    plt.show()
    # plt.close()  # Close the figure to avoid displaying it in notebooks or scripts


# Standard Library Imports
import time

# Third-Party Numerical and Scientific Libraries
import numpy as np
from scipy.stats import kde
import anndata
# from concurrent.futures import ProcessPoolExecutor


# Data Visualization Libraries
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

    
class SimData:
    def __init__(self, total_time=1, dt=0.005, K=10, n_vars=200, seed=0, 
                 a=None, beta=None, c=None, gamma=None, u0=None, s0=None):
        # Simulation Parameters
        self.total_time = total_time
        self.dt = dt
        self.K = K
        self.n_vars = a.shape[0] if a is not None else n_vars
        self.seed = seed

        # Model Parameters
        self.u0 = u0 if u0 is not None else 0.1*np.random.rand(self.n_vars)
        self.s0 = s0 if s0 is not None else 0.1*np.random.rand(self.n_vars)
        # self.u0 = 0.0
        # self.s0 = 0.0
        self.a = a if a is not None else 0.2 + 0.6 *np.random.rand(self.n_vars)
        # self.b = b if b is not None else 100.0
        self.c = c if c is not None else 2 + 10 *np.random.rand(self.n_vars)*np.random.rand(self.n_vars)
        self.beta = beta if beta is not None else 2 + 10 *np.random.rand(self.n_vars)
        self.gamma = gamma if gamma is not None else 2 + 10 *np.random.rand(self.n_vars)

        # Noise Parameters
        self.sigma_1 = 2.0 * np.random.rand(self.n_vars)
        self.sigma_2 = 2.0 * np.random.rand(self.n_vars)
        
        self.adata = None


    def generate(self):
        np.random.seed(self.seed if self.seed is not None else np.random.randint(10000))
        n_obs_per_trajectory = int(self.total_time / self.dt)
        n_obs = n_obs_per_trajectory * self.K

        U_layers = np.zeros((n_obs, self.n_vars))
        S_layers = np.zeros((n_obs, self.n_vars))

        for k in range(self.K):
            U, S = np.zeros((n_obs_per_trajectory, self.n_vars)), np.zeros((n_obs_per_trajectory, self.n_vars))
            Z1, Z2 = np.random.normal(0, 1, self.n_vars), np.random.normal(0, 1, self.n_vars)
            
            U[0] = np.maximum(0 + self.sigma_1 * np.sqrt(self.dt) * Z1, 0)
            S[0] = np.maximum(0 + self.sigma_2 * np.sqrt(self.dt) * Z2, 0)
            for t in range(n_obs_per_trajectory-1):
                time = t * self.dt
                alpha_t = self.c / (1 + np.exp(100.0 * (time - self.a)))
                # alpha_t = self.c / (1 + np.exp(100.0 * (time - 0.5)))
                Z1, Z2 = np.random.normal(0, 1, self.n_vars), np.random.normal(0, 1, self.n_vars)
                
                dU = (alpha_t - self.beta * U[t]) * self.dt + self.sigma_1 * np.sqrt(self.dt) * Z1
                dS = (self.beta * U[t] - self.gamma * S[t]) * self.dt + self.sigma_2 * np.sqrt(self.dt) * Z2
                
                U[t + 1] = np.maximum(U[t] + dU, 0)
                S[t + 1] = np.maximum(S[t] + dS, 0)
            U = self.u0 + U
            S = self.s0 + S

            start, end = k * n_obs_per_trajectory, (k + 1) * n_obs_per_trajectory
            U_layers[start: end], S_layers[start: end] = U, S

        obs = {"true_t": np.tile(np.arange(n_obs_per_trajectory) * self.dt, self.K)}
        var_params = {
            "true_t_": self.a, "true_alpha": self.c, "true_beta": self.beta, 
            "true_gamma": self.gamma, #"true_scaling": self.b, 
            "true_sigma_1": self.sigma_1, "true_sigma_2": self.sigma_2,
            "true_u0": self.u0, "true_s0": self.s0
        }
        layers = {"unspliced": U_layers, "spliced": S_layers}
        
        self.adata = anndata.AnnData(S_layers, obs=obs, var=var_params, layers=layers)
        return self.adata

    def plot_scatter(self, n1, n2, dpi=200):
        spliced = self.adata.layers['spliced']
        unspliced = self.adata.layers['unspliced']
        true_t = self.adata.obs['true_t']

        fig, axes = plt.subplots(n1, n2, figsize=(n2*4, n1*3), dpi=dpi)
        for i in range(n1 * n2):
            ax = axes[i//n2, i%n2] if (n1 > 1 or n2 > 1) else axes
            sc = ax.scatter(spliced[:, i], unspliced[:, i], c=true_t, s=30, alpha=0.5, cmap='viridis_r')
            plt.colorbar(sc, ax=ax)
            # ax.set_title(f'Scatter Plot for Variable {i}')
            ax.set_xlabel('Spliced')
            ax.set_ylabel('Unspliced')

        plt.tight_layout()
        plt.show()


from setuptools import setup, find_packages

setup(
    name="sdevelo",
    version="0.2.1",
    packages=find_packages(),
    install_requires=[
        'anndata',
        'matplotlib', 
        'numpy', 
        'scipy', 
        'scvelo>=0.3.0', 
        'scanpy',
        'seaborn', 
        'torch'
    ],
    author="Xu Liao",
    author_email="xu.smooth.liao@gmail.com",
    description="SDEvelo: a deep generative approach for transcriptional dynamics with cell-specific latent time and multivariate stochastic modeling",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/Liao-Xu/SDEvelo",
    license="LICENSE",
    classifiers=[
        # Classifiers help users find your project by categorizing it.
        # For a list of valid classifiers, see https://pypi.org/classifiers/
        "Development Status :: 1 - Planning", 
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)
